from __future__ import annotations

import functools
import typing
from typing import Any, ParamSpec

import anyio.to_thread
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

__all__ = ("UploadFile",)

P = ParamSpec("P")
T = typing.TypeVar("T")


async def run_in_thread_pool(
    func: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> T:
    if kwargs:  # pragma: no cover
        func = functools.partial(func, **kwargs)
    return await anyio.to_thread.run_sync(func, *args)


class UploadFile(bytes):
    def __new__(
        cls,
        file: typing.BinaryIO,
        *,
        size: int | None = None,
        filename: str | None = None,
        headers: dict[str, str] | None = None,
    ) -> bytes:
        obj = super().__new__(cls, file or b"")
        obj.filename = filename
        obj.size = size
        obj.headers = headers or {}
        obj.file = file
        return obj

    @property
    def content_type(self) -> str | None:
        return self.headers.get("content-type", None)

    def __len__(self) -> int:
        return len(self.file.read())

    def __getitem__(self, key) -> bytes:
        return self.file.read()[key]

    def __iter__(self) -> iter:
        return iter(self.file.read())

    def __bytes__(self) -> bytes:
        return self.file.read()

    def __instancecheck__(self, instance) -> bool:
        return isinstance(instance, bytes)

    def __subclasscheck__(self, subclass) -> bool:
        return issubclass(subclass, bytes)

    def __class__(self):
        return bytes

    @property
    def _in_memory(self) -> bool:
        # check for SpooledTemporaryFile._rolled
        rolled_to_disk = getattr(self.file, "_rolled", True)
        return not rolled_to_disk

    async def write(self, data: bytes) -> None:
        if self.size is not None:
            self.size += len(data)
        if self._in_memory:
            self.file.write(data)
        else:
            await run_in_thread_pool(self.file.write, data)

    async def read(self, size: int = -1) -> bytes:
        if self._in_memory:
            return self.file.read(size)
        return await run_in_thread_pool(self.file.read, size)

    async def seek(self, offset: int) -> None:
        if self._in_memory:
            self.file.seek(offset)
        else:
            await run_in_thread_pool(self.file.seek, offset)

    async def close(self) -> None:
        if self._in_memory:
            self.file.close()
        else:
            await run_in_thread_pool(self.file.close)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"filename={self.filename!r}, "
            f"size={self.size!r}, "
            f"headers={self.headers!r})"
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, _handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return {"type": "string", "format": "binary"}

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Any,
    ) -> core_schema.CoreSchema:
        return core_schema.bytes_schema()



import uuid

APP_GUARD = "__APP__GUARD__"
APP_FILTER = "__APP__FILTER__"
APP_INTERCEPTOR = "__APP__INTERCEPTOR__"


class _AppKey:
    @property
    def APP_GUARD(self):
        return APP_GUARD + uuid.uuid4().hex

    @property
    def APP_FILTER(self):
        return APP_FILTER + uuid.uuid4().hex

    @property
    def APP_INTERCEPTOR(self):
        return APP_INTERCEPTOR + uuid.uuid4().hex


AppKey = _AppKey()



from .decorator import UseFilters, Catch
from .http import HttpException
from .interface import ExceptionFilter
from .message import HttpStatusMessages
from .meta import ExceptionKey
from .status import HttpStatus

__all__ = [
    "HttpException",
    "HttpStatusMessages",
    "HttpStatus",
    "UseFilters",
    "Catch",
    "ExceptionKey",
    "ExceptionFilter",
]



from typing import Optional

from nestipy.ioc.dependency import Req, SocketData, Inject
from .context import RpcRequest

Payload = SocketData
Ctx = Req
Context = RpcRequest


def Client(token: Optional[str] = None):
    return Inject(token)


__all__ = ["Payload", "Ctx", "Client", "Context"]



from .event_module import EventEmitterModule
from .event_decorator import OnEvent, OnceEvent
from .event_emitter import EventEmitter

__all__ = ["EventEmitterModule", "OnEvent", "OnceEvent", "EventEmitter"]



class ModuleMetadata:
    Imports: str = "imports"
    Exports: str = "exports"
    Providers: str = "providers"
    Controllers: str = "controllers"
    Global: str = "_global_"
    Root: str = "_root_"
    Module: str = "_is_module_"



from typing import Union, Type, TYPE_CHECKING, Callable

from nestipy.metadata import SetMetadata

from nestipy.common.decorator import Injectable
from .meta import ExceptionKey

if TYPE_CHECKING:
    from .interface import ExceptionFilter
    from nestipy.common.exception.http import HttpException


def Catch(
    *exceptions: Union[Type["HttpException"], "HttpException"],
) -> Callable[[Type["ExceptionFilter"]], Type["ExceptionFilter"]]:
    """
    Catch decorator
    Args:
        *exceptions(Union[Type["HttpException"], "HttpException"]): List of exceptions

    Returns:
        decorator: Catch decorator
    """
    decorator = SetMetadata(ExceptionKey.MetaType, list(exceptions), as_list=True)

    def wrapper(cls: Type["ExceptionFilter"]):
        cls = Injectable()(cls)
        return decorator(cls)

    return wrapper


def UseFilters(*filters: Union[Type["ExceptionFilter"], "ExceptionFilter"]):
    """
    Filter decorator
    Args:
        *filters (Union[Type["ExceptionFilter"], "ExceptionFilter"]): List of filters

    Returns:
        decorator: Filter decorator
    """
    return SetMetadata(ExceptionKey.MetaFilter, list(filters), as_list=True)



from .openapi_docs.v3 import Info, OpenAPI, Components, SecurityScheme, HTTPSecurity


class DocumentBuilder:
    _info: dict = {"title": "Nestipy", "description": "", "version": "1.0"}
    _security_schemes: dict = {}
    _schemas: dict = {}
    _instance: "DocumentBuilder" = None

    @classmethod
    def get_instance(cls):
        return DocumentBuilder()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(DocumentBuilder, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def set_title(self, title: str):
        self._info["title"] = title
        return self

    def set_description(self, description: str):
        self._info["description"] = description
        return self

    def set_version(self, version: str):
        self._info["version"] = version
        return self

    def add_basic_auth(self):
        self._security_schemes["basicAuth"] = HTTPSecurity(
            scheme="basic",
        )
        return self

    def add_bearer_auth(self):
        self._security_schemes["bearer"] = HTTPSecurity(
            scheme="bearer",
        )
        return self

    def add_security(self, name, security: SecurityScheme):
        self._security_schemes[name] = security
        return self

    def build(self):
        return OpenAPI(
            info=Info(**self._info),
            components=Components(
                security_schemes=self._security_schemes, schemas=self._schemas
            ),
        )



from typing import Type, Callable, Union, Optional, Any

from nestipy.common import Module
from nestipy.core.nestipy_factory import NestipyFactory
from nestipy.core.platform import BlackSheepApplication, FastApiApplication
from nestipy.dynamic_module import DynamicModule
from nestipy.ioc import NestipyContainer
from nestipy.ioc.provider import ModuleProviderDict
from .client import TestClient


class TestingModuleRef:
    __test__ = False

    def __init__(self, root_module: Type):
        self._root_module = root_module

    async def get(self, key: Any):
        await NestipyFactory.create(self._root_module).ready()
        return await NestipyContainer.get_instance().get(key)

    def create_nestipy_client(
        self,
        platform: Union[
            Type[BlackSheepApplication], Type[FastApiApplication]
        ] = FastApiApplication,
    ) -> TestClient:
        return TestClient(NestipyFactory[platform].create(self._root_module))


class Test:
    __test__ = False

    @staticmethod
    def create_testing_module(
        providers: Optional[list[Callable | ModuleProviderDict]] = None,
        controllers: Optional[list[Callable]] = None,
        imports: Optional[
            list[Union[Type, Callable, ModuleProviderDict, DynamicModule]]
        ] = None,
        exports: Optional[list[Union[Type, Callable, str]]] = None,
        is_global: bool = False,
    ) -> TestingModuleRef:
        @Module(
            providers=providers,
            controllers=controllers,
            imports=imports,
            exports=exports,
            is_global=is_global,
        )
        class _RootTestingModule:
            pass

        return TestingModuleRef(_RootTestingModule)



from typing import Type, Union, Callable, Any, Optional

from nestipy.metadata import Reflect

from nestipy.graphql.meta import NestipyGraphqlKey


class GraphQlDecorator:
    def __init__(self, graphql_type: str, value: Any = True, is_method: bool = True):
        self.graphql_type = graphql_type
        self.graphql_value = value
        self.is_method = is_method

    def __call__(self, return_type: Optional[Type] = None, *args, **kwargs):
        def wrapper(class_ref: Union[Type, Callable]):
            from nestipy.common.decorator import Injectable

            if not self.is_method:
                class_ref = Injectable()(class_ref)
            Reflect.set_metadata(class_ref, self.graphql_type, self.graphql_value)
            Reflect.set_metadata(class_ref, NestipyGraphqlKey.return_type, return_type)
            Reflect.set_metadata(class_ref, NestipyGraphqlKey.args, args)
            Reflect.set_metadata(class_ref, NestipyGraphqlKey.kwargs, kwargs)
            return class_ref

        return wrapper


def Resolver(of: Optional[Type] = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.resolver, is_method=False)(
        of, *args, **kwargs
    )


def Query(return_type: Optional[Type] = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.handler, NestipyGraphqlKey.query)(
        return_type, *args, **kwargs
    )


def ResolveField(
    return_type: Optional[Type] = None, name: Optional[str] = None, *args, **kwargs
):
    return GraphQlDecorator(
        NestipyGraphqlKey.handler, NestipyGraphqlKey.field_resolver, True
    )(return_type, *args, **{"name": name, **kwargs})


def Mutation(return_type: Optional[Type] = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.handler, NestipyGraphqlKey.mutation)(
        return_type, *args, **kwargs
    )


def Subscription(return_type: Optional[Type] = None, *args, **kwargs):
    return GraphQlDecorator(NestipyGraphqlKey.handler, NestipyGraphqlKey.subscription)(
        return_type, *args, **kwargs
    )



from .base import ClientProxy, MicroserviceOption
from .grpc.client import GrpcClientProxy
from .mqtt import MqttClientProxy
from .nats import NatsClientProxy
from .option import RabbitMQClientOption, GrpcClientOption
from .option import (
    TCPClientOption,
    RedisClientOption,
    NatsClientOption,
    MqttClientOption,
)
from .option import Transport
from .rabbitmq import RabbitMQClientProxy
from .redis import RedisClientProxy
from .tcp import TCPClientProxy


class ClientModuleFactory:
    @classmethod
    def create(cls, option: MicroserviceOption) -> ClientProxy:
        match option.transport:
            case Transport.TCP:
                if option.option is None:
                    option.option = TCPClientOption()
                return TCPClientProxy(
                    option,
                )
            case Transport.REDIS:
                if option.option is None:
                    option.option = RedisClientOption()
                return RedisClientProxy(
                    option,
                )
            case Transport.MQTT:
                if option.option is None:
                    option.option = MqttClientOption()
                return MqttClientProxy(
                    option,
                )
            case Transport.RABBITMQ:
                if option.option is None:
                    option.option = RabbitMQClientOption()
                return RabbitMQClientProxy(
                    option,
                )
            case Transport.NATS:
                if option.option is None:
                    option.option = NatsClientOption()
                return NatsClientProxy(
                    option,
                )

            case Transport.GRPC:
                if option.option is None:
                    option.option = GrpcClientOption()
                return GrpcClientProxy(
                    option,
                )
            case Transport.CUSTOM:
                if option.proxy is not None and isinstance(option.proxy, ClientProxy):
                    return option.proxy
                else:
                    raise Exception("Unknown Transport")
            case _:
                raise Exception("Unknown Transport")



from .base import ClientProxy, MicroserviceOption
from .factory import ClientModuleFactory
from .module import ClientsModule, ClientsConfig
from .mqtt import MqttClientProxy
from .nats import NatsClientProxy
from .rabbitmq import RabbitMQClientProxy
from .redis import RedisClientProxy
from .tcp import TCPClientProxy
from .grpc.client import GrpcClientProxy

__all__ = [
    "ClientProxy",
    "MicroserviceOption",
    "TCPClientProxy",
    "RedisClientProxy",
    "MqttClientProxy",
    "RabbitMQClientProxy",
    "NatsClientProxy",
    "ClientsModule",
    "ClientsConfig",
    "ClientModuleFactory",
    "GrpcClientProxy",
]



import inspect
from contextvars import ContextVar
from typing import Optional, Dict, Any, Callable

from nestipy.common import Injectable


@Injectable()
class AsyncLocalStorage:
    def __init__(self):
        # Create a ContextVar to store the local state
        self._store: ContextVar[Optional[Dict[str, Any]]] = ContextVar(
            "store", default=None
        )

    async def run(self, data: Dict[str, Any], func: Callable[..., Any]):
        """
        Async version of the `run` method.
        """
        token = self._store.set(data)
        try:
            if inspect.iscoroutinefunction(func):
                return await func()  # Execute the async function within this context
            else:
                return func()  # Execute the sync function within this context
        finally:
            self._store.reset(token)

    def get_store(self) -> Optional[Dict[str, Any]]:
        """
        Retrieve the current store.
        """
        return self._store.get()

    def set_value(self, key: str, value: Any):
        """
        Set a specific key-value pair in the current store.
        """
        store = self._store.get()
        if store is not None:
            store[key] = value
        else:
            raise RuntimeError("No active store found")

    def get_value(self, key: str, default: Any = None) -> Any:
        """
        Retrieve a specific value from the current store.
        """
        store = self._store.get()
        if store is not None and key in store:
            return store[key]
        return default
        # raise KeyError(f"No value found for key: {key}")



"""
This module provides functions to generate Markdown for OpenAPI Version 3.
"""

import copy
import os
import warnings
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable, List, Optional, Union

from openapidocs.logs import logger
from openapidocs.mk import read_dict, sort_dict
from openapidocs.mk.common import (
    DocumentsWriter,
    get_ref_type_name,
    is_array_schema,
    is_object_schema,
    is_reference,
)
from openapidocs.mk.contents import ContentWriter, FormContentWriter, JSONContentWriter
from openapidocs.mk.jinja import Jinja2DocumentsWriter, OutputStyle
from openapidocs.mk.texts import EnglishTexts, Texts
from openapidocs.mk.v3.examples import get_example_from_schema
from openapidocs.utils.source import read_from_source


def _can_simplify_json(content_type) -> bool:
    return "json" in content_type or content_type == "text/plain"


def _can_simplify_xml(content_type) -> bool:
    return "xml" in content_type


class ExpandContext:
    def __init__(self) -> None:
        self.expanded_refs: set = set()


@dataclass
class ContentExample:
    value: Any
    auto_generated: bool
    name: str = ""


def style_from_value(value: Union[int, str]) -> OutputStyle:
    if isinstance(value, int) or value.isdigit():
        return OutputStyle(int(value))
    try:
        return OutputStyle[value]
    except KeyError:
        raise ValueError(f"Invalid style: {value}")


class OpenAPIDocumentationHandlerError(Exception):
    """Base type for exceptions raised by the handler generating documentation."""


class OpenAPIFileNotFoundError(OpenAPIDocumentationHandlerError, FileNotFoundError):
    """
    Exception raised when a $ref property pointing to a file (to split OAD specification
    into multiple files) is not found.
    """

    def __init__(self, reference: str, attempted_path: Path) -> None:
        super().__init__(
            f"Cannot resolve the $ref source {reference} "
            f"Tried to read from path: {attempted_path}"
        )


class OpenAPIV3DocumentationHandler:
    """
    Class that produces documentation from OpenAPI Documentation V3.
    """

    simplifiable_types = {
        "application/json": _can_simplify_json,
        "application/xml": _can_simplify_xml,
    }

    content_writers: List[ContentWriter] = [
        JSONContentWriter(),
        FormContentWriter(),
    ]

    default_content_writer = JSONContentWriter()

    def __init__(
        self,
        doc,
        texts: Optional[Texts] = None,
        writer: Optional[DocumentsWriter] = None,
        style: Union[int, str] = 1,
        source: str = "",
    ) -> None:
        self._source = source
        self.texts = texts or EnglishTexts()
        self._writer = writer or Jinja2DocumentsWriter(
            __name__, views_style=style_from_value(style)
        )
        self.doc = self.normalize_data(copy.deepcopy(doc))

    @property
    def source(self) -> str:
        return self._source

    def normalize_data(self, data):
        """
        Applies corrections to the OpenAPI specification, to simplify its handling.

        This method also resolves references to different files, if the root is split
        into multiple files.

        ---
        Ref.
        An OpenAPI document MAY be made up of a single document or be divided into
        multiple, connected parts at the discretion of the user. In the latter case,
        $ref fields MUST be used in the specification to reference those parts as
        follows from the JSON Schema definitions.
        """
        if "components" not in data:
            data["components"] = {}

        return self._transform_data(
            data, Path(self.source).parent if self.source else Path.cwd()
        )

    def _transform_data(self, obj, source_path):
        if not isinstance(obj, dict):
            return obj

        if "$ref" in obj:
            return self._handle_obj_ref(obj, source_path)

        clone = {}

        for key, value in obj.items():
            if isinstance(value, list):
                clone[key] = [self._transform_data(item, source_path) for item in value]
            elif isinstance(value, dict):
                clone[key] = self._handle_obj_ref(value, source_path)
            else:
                clone[key] = self._transform_data(value, source_path)

        return clone

    def _handle_obj_ref(self, obj, source_path):
        """
        Handles a dictionary containing a $ref property, resolving the reference if it
        is to a file. This is used to read specification files when they are split into
        multiple items.
        """
        assert isinstance(obj, dict)
        if "$ref" in obj:
            reference = obj["$ref"]
            if isinstance(reference, str) and not reference.startswith("#/"):
                referred_file = Path(os.path.abspath(source_path / reference))

                if referred_file.exists():
                    logger.debug("Handling $ref source: %s", reference)
                else:
                    raise OpenAPIFileNotFoundError(reference, referred_file)
                sub_fragment = read_from_source(str(referred_file))
                return self._transform_data(sub_fragment, referred_file.parent)
            else:
                return obj
        return self._transform_data(obj, source_path)

    def get_operations(self):
        """
        Gets a dictionary of operations grouped by tag.
        """
        data = self.doc
        groups = defaultdict(list)
        paths = data["paths"]

        for path, path_item in paths.items():
            tag = self.get_tag(path_item) or ""

            for operation in path_item.values():
                # need to resolve possible references for requestBody
                if "requestBody" in operation:
                    operation["requestBody"] = self._resolve_opt_ref(
                        operation["requestBody"]
                    )

            groups[tag].append((path, path_item))

        return groups

    def get_schemas(self):
        schemas = read_dict(self.doc, "components", "schemas")

        if not schemas:
            return

        yield from sort_dict(schemas)

    def get_tag(self, path_item) -> Optional[str]:
        """
        Tries to obtain a single tag for all operations inside a given path item.
        See https://spec.openapis.org/oas/v3.1.0#path-item-object

        A path item looks like this:

        {
            "get": {..., "tags": ["Albums"]},
            "post": {..., "tags": ["Albums"]}
        }

        Tags are optional.
        """
        single_tag: Optional[str] = None

        for operation in path_item.values():
            tags = operation.get("tags")

            if not tags:
                continue

            single_tag = next(
                (item for item in tags if single_tag is None or item == single_tag),
                None,
            )

        return single_tag

    def simplify_content(self, content):
        """
        Gets a copy of the content definition, altered to be more concise, when several
        content types can be used to describe the same schema, returning a dictionary of
        references and possible content-types.

        Example:

        {
            "text/plain": {
                "schema": {
                    "$ref": "#/components/schemas/Release"
                }
            },
            "application/json": {
                "schema": {
                    "$ref": "#/components/schemas/Release"
                }
            },
            "text/json": {
                "schema": {
                    "$ref": "#/components/schemas/Release"
                }
            }
        }

        ->

        {
            "application/json": {
                "schema": {
                    "$ref": "#/components/schemas/Release"
                },
                "alt_types": ["text/plain", "text/json"]
            }
        }
        """
        simplified_content = copy.deepcopy(content)
        all_types = simplified_content.keys()

        for content_type, predicate in self.simplifiable_types.items():
            main_declaration = simplified_content.get(content_type)

            if main_declaration:
                types_to_remove = {
                    other_type
                    for other_type in all_types
                    if predicate(other_type)
                    and other_type != content_type
                    and simplified_content[other_type] == main_declaration
                }

                for type_to_remove in types_to_remove:
                    del simplified_content[type_to_remove]

                if types_to_remove:
                    alt_types = list(types_to_remove)
                    # sort to use alphabetically sorted values
                    alt_types.sort()
                    main_declaration["alt_types"] = alt_types

        return simplified_content

    def get_security_scheme(self, name: str) -> dict:
        """
        Gets a security scheme from the components section, by name.
        """
        security_scheme = read_dict(self.doc, "components", "securitySchemes")

        if not security_scheme:  # pragma: no cover
            warnings.warn(
                "Missing section in components. A security scheme referenced "
                "in a path item is not configured."
            )
            return {}

        if name not in security_scheme:  # pragma: no cover
            warnings.warn(
                f'Missing security scheme definition "{name}". '
                "A path item references this scheme, but it is not configured "
                "in components.securitySchemes."
            )
            return {}
        return security_scheme[name]

    def get_parameter_for_security(self, key: str, security_scheme: dict):
        """
        The OpenAPI Documentation specification is messy. The security scheme objects
        describe input parameters in a completely different way than input parameters,
        and try to cover too many scenarios (OAuth flows).

        https://swagger.io/specification/#security-scheme-object
        https://swagger.io/specification/#security-requirement-object
        """
        security_type = security_scheme.get("type")

        if security_type == "http":
            scheme = security_scheme.get("scheme")

            description = ""
            if scheme == "bearer":
                description = "JWT Bearer token"
            elif scheme == "basic":
                description = "Basic authentication"

            return {
                "name": key,
                "in": "header",
                "description": security_scheme.get("description", description),
                "schema": {"type": "string", "default": "N/A", "nullable": False},
            }
        if security_type == "apiKey":
            return {
                "name": key,
                "in": security_scheme.get("in"),
                "description": security_scheme.get("description", "API key"),
                "schema": {"type": "string", "default": "N/A", "nullable": False},
            }

        return {
            "name": key,
            "in": "header",
            "description": security_scheme.get("description", ""),
            "schema": {
                "type": "string",
                "default": "N/A",
                "nullable": False,
            },
        }

    def get_operation_security(self, operation):
        """
        Returns security definition for an operation. This can come from global settings
        or from specific settings.

        Example:
            security:
                - ApiKeyAuth: []
                - OAuth2: [read, write]

            Refer to:

            https://swagger.io/specification/#security-scheme-object
            https://swagger.io/specification/#security-requirement-object
        """
        for source in [operation, self.doc]:
            if "security" in source:
                return source["security"]
        return None

    def _resolve_opt_ref(self, obj):
        if "$ref" in obj:
            return self.resolve_reference(obj)
        return obj

    def get_parameters(self, operation) -> List[dict]:
        """
        Returns a list of objects describing the input parameters for a given operation.
        References to #/components/parameters are resolved, to show the information in
        a single place.
        """
        parameters = [
            self._resolve_opt_ref(item) for item in operation.get("parameters", [])
        ]

        results = [
            param
            for param in sorted(
                parameters,
                key=lambda x: x["name"].lower() if (x and "name" in x) else "",
            )
            if param
        ]

        security_options = self.get_operation_security(operation)
        if security_options:
            # The OAD specification here is messy and confusing: it treats input
            # parameters for authentication like they were a completely different thing
            # than other input parameters - which in most cases is not true
            # (Authorization headers, cookies, API Keys in headers, etc.).

            # Here we prepend the parameter for authentication | authorization,
            # so that readers who are not expert of various authentication methods can
            # see that a certain header / parameter is needed.
            assert isinstance(security_options, list)
            for option in security_options:
                for key in option.keys():
                    # TODO: support for showing scopes
                    scheme = self.get_security_scheme(key)
                    results.insert(0, self.get_parameter_for_security(key, scheme))

        return results

    def write(self) -> str:
        return self._writer.write(
            self.doc,
            operations=self.get_operations(),
            texts=self.texts,
            handler=self,
        )

    def get_content_examples(self, data) -> Iterable[ContentExample]:
        """
        Returns examples to show an example for a content definition.
        The specification allows several options: examples, example, objects, raw
        strings. If no example is specified, this method generates one automatically
        from the schema definition.

        Example (YAML):

        requestBody:
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/CreateCatInput'
                    examples:
                        fat_cat:
                            value:
                                name: Fatty
                                active: false
                                type: european
                        thin_cat:
                            value:
                                name: Thinny
                                active: false
                                type: persian
            required: true
            description: Example description etc. etc.


        responses:
            '200':
                description: A cat
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/Cat'
                        example:
                            id: 3fa85f64-5717-4562-b3fc-2c963f66afa6
                            name: Foo
                            active: true
                            type: european
                            creation_time: '2020-10-25T19:39:31.751652'
        """
        examples = data.get("examples")

        if examples:
            for example_name, definition in examples.items():
                if isinstance(definition, str):
                    value = definition
                elif is_reference(definition):
                    value = self.resolve_reference(definition)
                else:
                    value = definition.get("value")
                yield ContentExample(value, False, example_name)

        example = data.get("example")
        auto_generated = False

        if not example and not examples:
            # try to generate a basic example from the schema, on the fly
            example = self.generate_example_from_schema(data.get("schema"))
            auto_generated = True

        if example:
            yield ContentExample(example, auto_generated)

        return ""

    def write_content_example(self, example: ContentExample, content_type: str) -> str:
        example_handler = self.get_content_writer(content_type)
        return example_handler.write(example.value)

    def generate_example_from_schema(self, schema) -> Any:
        if not schema:
            return None

        if is_reference(schema):
            schema = self.resolve_reference(schema["$ref"])

        return get_example_from_schema(self.expand_references(schema))

    def write_content_schema(self, data) -> str:
        schema = data.get("schema")

        if not schema:
            return ""

        if is_reference(schema):
            schema = self.resolve_reference(schema["$ref"])

        return self.default_content_writer.write(schema)

    def resolve_reference(self, reference: Union[str, dict]) -> Any:
        """
        Returns the schema object from the components, by reference ($ref).
        """
        if isinstance(reference, dict):
            reference = reference["$ref"]
        assert isinstance(reference, str)
        return read_dict(self.doc, *reference.lstrip("#/").split("/"))

    def expand_references(self, schema, context: Optional[ExpandContext] = None):
        """
        Returns a clone of the given schema object, in which all $ref properties are
        resolved and made verbose. This is to provide a better view of schemas that does
        not require navigating to different parts of the documentation.

        This method handles recursive references setting `null` values.
        """
        if context is None:
            context = ExpandContext()

        if is_reference(schema):
            return self.expand_references(self.resolve_reference(schema))

        if schema is None:
            # this should not happen, but we don't want the whole build to fail
            return None

        clone = copy.deepcopy(schema)

        for key in list(clone.keys()):
            value = clone[key]

            if is_reference(value):
                ref = value["$ref"]
                if ref in context.expanded_refs:
                    clone[key] = None
                else:
                    context.expanded_refs.add(ref)

                    resolved_ref = self.resolve_reference(value)

                    if resolved_ref is None:  # pragma: no cover
                        logger.warning(
                            "Cannot resolve the reference %s. "
                            "Is a fragment missing from `components` object?",
                            value,
                        )
                        clone[key] = {}
                    else:
                        clone[key] = self.expand_references(resolved_ref, context)
            elif isinstance(value, dict):
                if is_array_schema(value) and is_reference(value["items"]):
                    ref = value["items"]["$ref"]
                    if ref in context.expanded_refs:
                        clone[key] = None
                        continue

                clone[key] = self.expand_references(value, context)
            else:
                clone[key] = value

        return clone

    def get_content_writer(self, content_type: str) -> ContentWriter:
        """
        Returns a ContentWriter to create a markdown representation of the given
        content type. If none specific is available, it returns the
        `default_content_writer` bound to this class.
        """
        return next(
            (
                writer
                for writer in self.content_writers
                if writer.handle_content_type(content_type)
            ),
            self.default_content_writer,
        )

    def get_properties(self, schema):
        if is_reference(schema):
            schema = self.resolve_reference(schema)

            if not schema:
                return []

        return [[key, value] for key, value in sort_dict(schema.get("properties", {}))]

    def iter_schemas_bindings(self):
        """
        Iterates through components schemas and yields tuples of type names and their
        $refs types.
        """
        schemas = self.get_schemas()

        for type_name, schema in schemas:
            if is_object_schema(schema):
                for _, prop_schema in sort_dict(schema["properties"]):
                    if is_reference(prop_schema):
                        yield type_name, get_ref_type_name(prop_schema)

                    if is_array_schema(prop_schema) and is_reference(
                        prop_schema["items"]
                    ):
                        yield type_name, get_ref_type_name(prop_schema["items"])

            if is_array_schema(schema) and is_reference(schema["items"]):
                yield type_name, get_ref_type_name(schema["items"])



from typing import Annotated

from nestipy.common import Module
from nestipy.core import DiscoverService
from nestipy.dynamic_module import NestipyModule
from nestipy.ioc import Inject
from nestipy.metadata import RouteKey, Reflect
from .router_builder import ConfigurableClassBuilder, ROUTER_CONFIG, RouteItem


@Module()
class RouterModule(ConfigurableClassBuilder, NestipyModule):
    _option: Annotated[list[RouteItem], Inject(ROUTER_CONFIG)]
    _discover: Annotated[DiscoverService, Inject()]

    async def on_startup(self):
        self._update_router([], self._option)

    def _update_router(self, main_path: [], items: list[RouteItem]):
        for item in items:
            parent_path = [*main_path]
            controllers = self._discover.get_module_controllers(item.module)
            for controller in controllers:
                controller_path = self._normalize_path(
                    Reflect.get_metadata(controller, RouteKey.path, "")
                )
                Reflect.set_metadata(
                    controller,
                    RouteKey.path,
                    "/".join(
                        [self._normalize_path(p) for p in parent_path]
                        + [self._normalize_path(item.path), controller_path]
                    ),
                )
            parent_path.append(item.path)
            if item.children:
                self._update_router(parent_path, item.children)

    @classmethod
    def register(cls, config: list[RouteItem]):
        return cls._register(config)

    @classmethod
    def _normalize_path(cls, path: str) -> str:
        return path.strip("/")



from abc import ABC, abstractmethod
from typing import Any, Optional

import click


class BaseCommand(ABC):
    """Abstract base class for Nestipy commands."""

    _args: list = []
    _options: dict[str, Any] = {}

    @abstractmethod
    async def run(self):
        """The method to handle the command logic."""
        raise NotImplemented("Must be implement")

    @classmethod
    def info(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="green")

    @classmethod
    def error(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="red")

    @classmethod
    def warning(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="orange")

    @classmethod
    def success(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="green", bold=True)

    def init(self, context: tuple):
        """Parse the combined context into arguments and options.

        Args:
            context (tuple): Combined arguments and options.

        """
        positional_args = []
        keyword_args = {}
        i = 0
        while i < len(context):
            arg = context[i]
            if arg.startswith("--"):
                if "=" in arg:
                    key, value = arg[2:].split("=", 1)
                    keyword_args[key] = value
                else:
                    key = arg[2:]
                    # lookahead for value
                    if i + 1 < len(context) and not context[i + 1].startswith("-"):
                        keyword_args[key] = context[i + 1]
                        i += 1
                    else:
                        keyword_args[key] = True
            elif arg.startswith("-") and len(arg) == 2:
                key = arg[1:]
                if i + 1 < len(context) and not context[i + 1].startswith("-"):
                    keyword_args[key] = context[i + 1]
                    i += 1
                else:
                    keyword_args[key] = True
            else:
                positional_args.append(arg)
            i += 1

        self._args = positional_args
        self._options = keyword_args

    def get_opt(self, key: str = None, default=None):
        """Retrieve an option from the options dictionary safely."""
        return self._options.get(key, default) if key else self._options

    def get_arg(self, index: int = None, default=None):
        """Retrieve an argument by index from the arguments list safely."""
        try:
            return self._args[index] if index is not None else self._args
        except IndexError:
            return default



from typing import Type

from nestipy.ioc import ModuleProviderDict
from nestipy.metadata import Reflect, ClassMetadata, ModuleMetadata


class SpecialProviderExtractor:
    @classmethod
    def extract_special_providers(
        cls, module_class: Type, subclass: Type, key: str
    ) -> list[Type]:
        providers = []
        class_providers = Reflect.get_metadata(
            module_class, ModuleMetadata.Providers, []
        )
        metadata: ClassMetadata = Reflect.get_metadata(
            module_class, ClassMetadata.Metadata, None
        )
        if metadata is not None:
            class_providers += metadata.get_global_providers()
        for p in class_providers:
            if (
                isinstance(p, ModuleProviderDict)
                and isinstance(p.token, str)
                and p.token.startswith(key)
                and p.use_class is not None
            ):
                if issubclass(p.use_class, subclass):
                    providers.append(p.use_class)
        return providers



import os
from abc import abstractmethod
from typing import Callable, cast, Optional

import aiofiles

from nestipy.ioc import RequestContextContainer, NestipyContainer

from nestipy.common.http_ import Request, Response
from nestipy.graphql.graphql_module import GraphqlOption


class GraphqlASGI:
    option: Optional[GraphqlOption] = None

    @abstractmethod
    def print_schema(self) -> str:
        pass

    def set_graphql_option(self, option: GraphqlOption):
        self.option = option

    async def get_graphql_ide(self) -> str:
        playground_path = os.path.join(
            os.path.dirname(__file__),
            "playground",
            f"graphql-playground-{self.option.ide}.html",
        )
        file = await aiofiles.open(playground_path, "r")
        content = await file.read()
        await file.close()
        return content

    @abstractmethod
    async def handle(self, scope: dict, receive, send):
        self._setup_request(scope, receive, send)

    async def modify_default_context(self, context: dict) -> dict:
        if self.option.context_callback:
            return await NestipyContainer.get_instance().resolve_factory(
                self.option.context_callback,
                inject=[],
                search_scope=[],
                disable_scope=True,
            )
        return context

    @classmethod
    def _setup_request(cls, scope: dict, receive, send):
        from nestipy.core import ExecutionContext

        req_container = RequestContextContainer.get_instance()
        req = Request(scope, receive, send)
        res = Response()
        req_container.set_execution_context(
            ExecutionContext(
                None, None, None, cast(Callable, None), req, res, None, None
            )
        )



import asyncio

from nestipy.microservice import NatsClientOption, Transport
from nestipy.microservice.client import NatsClientProxy, MicroserviceOption
from nestipy.microservice.server.base import MicroServiceServer


async def main():
    client = NatsClientProxy(
        MicroserviceOption(transport=Transport.NATS, option=NatsClientOption()),
    )
    server = MicroServiceServer(pub_sub=client)
    print("Running server ...")
    await server.listen()


if __name__ == "__main__":
    asyncio.run(main())



import importlib.metadata
import os
import platform
from dataclasses import dataclass, field
from typing import List
from typing import Union


@dataclass
class FrameworkTrack:
    python: str
    nestipy: str


@dataclass
class RequestTrack:
    method: str
    host: str


@dataclass
class Traceback:
    filename: str
    lineno: int
    name: str
    code: str
    start_line_number: int = field(default=1)
    is_package: bool = field(default=False)


@dataclass
class ExceptionDetail:
    exception: str
    type: str
    message: str
    root: str
    request: RequestTrack = field(
        default_factory=lambda: RequestTrack("GET", "0.0.0.0:8000")
    )
    traceback: List[Traceback] = field(default_factory=lambda: [])
    framework: FrameworkTrack = field(
        default_factory=lambda: FrameworkTrack(
            platform.python_version(), importlib.metadata.version("nestipy")
        )
    )


class HttpException(Exception):
    def __init__(
        self,
        status_code: int,
        message: Union[str, None] = None,
        details: Union[dict, str, None] = None,
        track_back: Union[ExceptionDetail, None] = None,
    ):
        self.status_code = status_code
        self.message = message
        self.details = details
        self.track_back = track_back or ExceptionDetail(
            "Internal server error",
            "HttpException",
            message,
            os.getcwd(),
        )
        super().__init__(self.message)

    def __str__(self):
        return f"{self.status_code} - {self.message} {f' - {self.details}' if self.details is not None else ''}"



import asyncio
import typing
from dataclasses import asdict
from typing import AsyncIterator

from redis.asyncio import Redis, client

from .base import ClientProxy, MicroserviceOption
from .option import RedisClientOption


class RedisClientProxy(ClientProxy):
    broker: Redis
    pub_sub: client.PubSub

    def __init__(self, option: MicroserviceOption):
        super().__init__(option)

    async def connect(self):
        option = typing.cast(
            RedisClientOption, self.option.option or RedisClientOption()
        )
        # url = f"redis://{option.host}:{option.port}"
        # self.broker = Redis.from_url(url, max_connections=10, decode_responses=True)
        self.broker = Redis(**asdict(option))
        self.pub_sub = self.broker.pubsub()
        await self.pub_sub.connect()

    async def _publish(self, topic, data):
        await self.broker.publish(topic, data)

    async def send_response(self, topic, data):
        await self._publish(topic, data)

    async def listen(self) -> AsyncIterator[str]:
        while True:
            message = await self.pub_sub.get_message(ignore_subscribe_messages=True)
            if message is not None:
                if message["type"] == "message":
                    yield message["data"]
            await asyncio.sleep(0.01)

    async def listen_response(self, from_topic: str, timeout: int = 30) -> str:
        async for message in self.pub_sub.listen():
            if message is not None:
                if message["type"] == "message":
                    response = message["data"]
                    return response

    async def close(self):
        await self.pub_sub.aclose()
        await self.broker.close()

    async def subscribe(self, *args, **kwargs):
        await self.pub_sub.subscribe(*args, **kwargs)

    async def unsubscribe(self, *args):
        await self.pub_sub.unsubscribe(*args)

    async def slave(self) -> "ClientProxy":
        return RedisClientProxy(self.option)



from abc import ABC, abstractmethod
from typing import Any, Callable, Optional

from socketio import AsyncServer

from .socket_request import Websocket


class IoAdapter(ABC):
    def __init__(self, path: str = "socket.io"):
        self._path = f"/{path.strip('/')}"

    @abstractmethod
    def on(
        self, event: str, namespace: Optional[str] = None
    ) -> Callable[[Callable], Any]:
        pass

    @abstractmethod
    def emit(
        self,
        event: Any,
        data: Optional[Any] = None,
        to: Optional[Any] = None,
        room: Optional[Any] = None,
        skip_sid: Optional[Any] = None,
        namespace: Optional[Any] = None,
        callback: Optional[Any] = None,
        ignore_queue: bool = False,
    ):
        pass

    @abstractmethod
    def on_connect(self) -> Callable[[Callable], Any]:
        pass

    @abstractmethod
    def on_disconnect(self) -> Callable[[Callable], Any]:
        pass

    @abstractmethod
    def broadcast(self, event: Any, data: Any):
        pass

    @abstractmethod
    async def __call__(self, scope: dict, receive: Callable, send: Callable) -> bool:
        pass


class SocketIoAdapter(IoAdapter):
    def __init__(self, io: AsyncServer, path: str = "socket.io"):
        super().__init__(path=path)
        self._io: AsyncServer = io
        self._connected: list = []

    def on(self, event: str, namespace: str = None):
        def decorator(handler: Callable):
            async def wrapper(sid: str, data: Any):
                environ = self._io.get_environ(sid, namespace)
                client = Websocket(
                    namespace,
                    sid,
                    data,
                    environ["asgi.scope"],
                    environ["asgi.receive"],
                    environ["asgi.send"],
                )
                return await handler(event, client, data)

            self._io.on(event, namespace=namespace)(wrapper)

        return decorator

    async def emit(
        self,
        event: Any,
        data: Any = None,
        to: Any = None,
        room: Any = None,
        skip_sid: Any = None,
        namespace: Any = None,
        callback: Any = None,
        ignore_queue: bool = False,
    ):
        return await self._io.emit(
            event, data, to, room, skip_sid, namespace, callback, ignore_queue
        )

    def broadcast(self, event: Any, data: Any):
        return self._io.emit(event, data, self._connected)

    def on_connect(self):
        def decorator(handler: Callable):
            async def wrapper(sid: Any, *args, **kwargs):
                self._connected.append(sid)
                return await handler(sid, *args, **kwargs)

            return self._io.on("connect")(wrapper)

        return decorator

    def on_disconnect(self):
        def decorator(handler: Callable):
            async def wrapper(sid: Any, *args, **kwargs):
                self._connected.remove(sid)
                return await handler(sid, *args, **kwargs)

            return self._io.on("disconnect")(wrapper)

        return decorator

    async def __call__(self, scope: dict, receive: Callable, send: Callable):
        if scope["type"] in ["http", "websocket"] and scope["path"].startswith(
            self._path
        ):
            await self._io.handle_request(scope, receive, send)
            return True
        return False



import dataclasses
from datetime import timedelta
from typing import Literal, Annotated, Optional, Sequence, Callable

from nestipy.common.decorator import Module
from nestipy.dynamic_module import ConfigurableModuleBuilder
from nestipy.ioc import Inject


@dataclasses.dataclass
class ASGIOption:
    allow_queries_via_get: bool = True
    keep_alive: bool = False
    keep_alive_interval: float = 1
    debug: bool = False
    subscription_protocols: Sequence[str] = None
    connection_init_wait_timeout: timedelta = timedelta(minutes=1)


@dataclasses.dataclass
class GraphqlOption:
    url: str = "/graphql"
    cors: bool = True
    auto_schema_file: Optional[str] = None
    ide: Literal["default", "graphiql"] = "default"
    schema_option: Optional[dict] = None
    asgi_option: Optional[ASGIOption] = None
    context_callback: Optional[Callable[[...], dict]] = None

    def __post_init__(self):
        if self.ide not in ["default", "graphiql"]:
            raise ValueError("ide value must be one of 'default', 'graphiql'")


ConfigurableModuleClass, CONFIG_MODULE_OPTION_TOKEN = (
    ConfigurableModuleBuilder[GraphqlOption]().set_method("for_root").build()
)


@Module()
class GraphqlModule(ConfigurableModuleClass):
    config: Annotated[GraphqlOption, Inject(CONFIG_MODULE_OPTION_TOKEN)]



from dataclasses import dataclass
from typing import Any, Union

from nestipy.metadata import SetMetadata
from .client.option import Transport

MICROSERVICE_LISTENER = "__Microservice__Listener__"


class MicroserviceMetadata:
    Event = "__Microservice__Event__"
    Message = "__Microservice__Message__"


@dataclass
class MicroserviceData:
    pattern: Any
    type: str
    transport: Union[Transport, None] = None


def MessagePattern(pattern: Any, transport: Union[Transport, None] = None):
    return SetMetadata(
        MICROSERVICE_LISTENER,
        MicroserviceData(pattern, MicroserviceMetadata.Message, transport),
    )


def EventPattern(pattern: Any, transport: Union[Transport, None] = None):
    return SetMetadata(
        MICROSERVICE_LISTENER,
        MicroserviceData(pattern, MicroserviceMetadata.Event, transport),
    )



from abc import abstractmethod, ABC
from typing import Awaitable, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from nestipy.core.context.execution_context import ExecutionContext


class CanActivate(ABC):
    @abstractmethod
    def can_activate(self, context: "ExecutionContext") -> Union[Awaitable[bool], bool]:
        """
        Method to handle context and validate it by returning boolean or None
        Args:
            context(ExecutionContext): ExecutionContext of request
        """
        pass



from openapi_docs.common import Serializer
from openapi_docs.v3 import (
    OpenAPI,
    Info,
    Response,
    Components,
    Schema,
    SecuritySchemeType,
    MediaType,
)
from openapi_docs.v3 import PathItem, Operation, SecurityRequirement, ValueType
from openapi_docs.v3 import (
    Tag,
    HTTPSecurity,
    APIKeySecurity,
    ParameterLocation,
    Parameter,
    RequestBody,
)

open_api = OpenAPI(
    info=Info(
        title="Test openapi", description="This is the description", version="v1"
    ),
    paths={
        "/test": PathItem(
            summary="Test",
            get=Operation(
                responses={"200": Response(description="Response successfully")},
                tags=["Test"],
                operation_id="test",
                security=[SecurityRequirement(name="Bearer", value=["bearer"])],
                parameters=[
                    Parameter(name="test", in_=ParameterLocation.PATH),
                    Parameter(name="test2", in_=ParameterLocation.QUERY),
                    Parameter(name="test3", in_=ParameterLocation.HEADER),
                    Parameter(name="test3", in_=ParameterLocation.HEADER),
                ],
                request_body=RequestBody(
                    required=True,
                    content={
                        "application/json": MediaType(
                            schema=Schema(ref="#/components/schemas/User")
                        )
                    },
                ),
            ),
        )
    },
    components=Components(
        schemas={
            "User": Schema(
                type=ValueType.OBJECT,
                properties={
                    "id": Schema(type=ValueType.INTEGER, example=1, default=1),
                    "name": Schema(
                        type=ValueType.STRING, example="name", default="name"
                    ),
                },
            )
        },
        security_schemes={
            "bearer": HTTPSecurity(
                scheme="bearer",
            ),
            "apiKey": APIKeySecurity(name="X-API", in_=ParameterLocation.HEADER),
            "basicAuth": HTTPSecurity(scheme="basicAuth", type=SecuritySchemeType.HTTP),
        },
    ),
    tags=[Tag(name="Test"), Tag(name="Test2")],
)

serializer = Serializer()
if __name__ == "__main__":
    print(serializer.to_json(open_api))



from .annotation import ParamAnnotation
from .container import NestipyContainer
from .context_container import RequestContextContainer
from .dependency import (
    Inject,
    Res,
    Req,
    Session,
    Query,
    Body,
    Arg,
    Context,
    GraphQlContext,
    WebSocketServer,
    WebSocketClient,
    SocketData,
    Param,
    Header,
    Cookie,
    create_type_annotated,
)
from .middleware import MiddlewareProxy, MiddlewareRouteConfig, MiddlewareContainer
from .provider import ModuleProviderDict

__all__ = [
    "ParamAnnotation",
    "NestipyContainer",
    "RequestContextContainer",
    "ModuleProviderDict",
    "Inject",
    "Res",
    "Req",
    "Session",
    "Query",
    "Body",
    "Arg",
    "Context",
    "WebSocketServer",
    "WebSocketClient",
    "SocketData",
    "Param",
    "MiddlewareProxy",
    "MiddlewareRouteConfig",
    "MiddlewareContainer",
    "Header",
    "Cookie",
    "GraphQlContext",
    "create_type_annotated",
]
